var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};

var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

var Rule = function () {
  function Rule(name, fn, args, modifiers) {
    classCallCheck(this, Rule);

    this.name = name;
    this.fn = fn;
    this.args = args;
    this.modifiers = modifiers;
  }

  createClass(Rule, [{
    key: "_test",
    value: function _test(value) {
      var fn = this.fn;

      try {
        fn(value);
      } catch (ex) {
        fn = function fn() {
          return false;
        };
      }

      try {
        return testAux(this.modifiers.slice(), fn)(value);
      } catch (ex) {
        return false;
      }
    }
  }, {
    key: "_check",
    value: function _check(value) {
      try {
        this.fn(value);
      } catch (ex) {
        if (testAux(this.modifiers.slice(), function (it) {
          return it;
        })(false)) {
          return;
        }
      }

      if (!testAux(this.modifiers.slice(), this.fn)(value)) {
        throw null;
      }
    }
  }, {
    key: "_testAsync",
    value: function _testAsync(value) {
      var _this = this;

      return new Promise(function (resolve, reject) {
        testAsyncAux(_this.modifiers.slice(), _this.fn)(value).then(function (valid) {
          if (valid) {
            resolve(value);
          } else {
            reject(null);
          }
        }).catch(function (ex) {
          return reject(ex);
        });
      });
    }
  }]);
  return Rule;
}();

function testAux(modifiers, fn) {
  if (modifiers.length) {
    var modifier = modifiers.shift();
    var nextFn = testAux(modifiers, fn);
    return modifier.perform(nextFn);
  } else {
    return fn;
  }
}

function testAsyncAux(modifiers, fn) {
  if (modifiers.length) {
    var modifier = modifiers.shift();
    var nextFn = testAsyncAux(modifiers, fn);
    return modifier.performAsync(nextFn);
  } else {
    return function (value) {
      return Promise.resolve(fn(value));
    };
  }
}

var Modifier = function Modifier(name, perform, performAsync) {
  classCallCheck(this, Modifier);

  this.name = name;
  this.perform = perform;
  this.performAsync = performAsync;
};

var ValidationError = function (_Error) {
  inherits(ValidationError, _Error);

  function ValidationError(rule, value, cause, target) {
    classCallCheck(this, ValidationError);

    for (var _len = arguments.length, remaining = Array(_len > 4 ? _len - 4 : 0), _key = 4; _key < _len; _key++) {
      remaining[_key - 4] = arguments[_key];
    }

    var _this = possibleConstructorReturn(this, (ValidationError.__proto__ || Object.getPrototypeOf(ValidationError)).call(this, remaining));

    if (Error.captureStackTrace) {
      Error.captureStackTrace(_this, ValidationError);
    }
    _this.rule = rule;
    _this.value = value;
    _this.cause = cause;
    _this.target = target;
    return _this;
  }

  return ValidationError;
}(Error);

var Context = function () {
  function Context() {
    var chain = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
    var nextRuleModifiers = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    classCallCheck(this, Context);

    this.chain = chain;
    this.nextRuleModifiers = nextRuleModifiers;
  }

  createClass(Context, [{
    key: "_applyRule",
    value: function _applyRule(ruleFn, name) {
      var _this = this;

      return function () {
        for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
          args[_key] = arguments[_key];
        }

        _this.chain.push(new Rule(name, ruleFn.apply(_this, args), args, _this.nextRuleModifiers));
        _this.nextRuleModifiers = [];
        return _this;
      };
    }
  }, {
    key: "_applyModifier",
    value: function _applyModifier(modifier, name) {
      this.nextRuleModifiers.push(new Modifier(name, modifier.simple, modifier.async));
      return this;
    }
  }, {
    key: "_clone",
    value: function _clone() {
      return new Context(this.chain.slice(), this.nextRuleModifiers.slice());
    }
  }, {
    key: "test",
    value: function test(value) {
      return this.chain.every(function (rule) {
        return rule._test(value);
      });
    }
  }, {
    key: "testAll",
    value: function testAll(value) {
      var err = [];
      this.chain.forEach(function (rule) {
        try {
          rule._check(value);
        } catch (ex) {
          err.push(new ValidationError(rule, value, ex));
        }
      });
      return err;
    }
  }, {
    key: "check",
    value: function check(value) {
      this.chain.forEach(function (rule) {
        try {
          rule._check(value);
        } catch (ex) {
          throw new ValidationError(rule, value, ex);
        }
      });
    }
  }, {
    key: "testAsync",
    value: function testAsync(value) {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        executeAsyncRules(value, _this2.chain.slice(), resolve, reject);
      });
    }
  }]);
  return Context;
}();

function executeAsyncRules(value, rules, resolve, reject) {
  if (rules.length) {
    var rule = rules.shift();
    rule._testAsync(value).then(function () {
      executeAsyncRules(value, rules, resolve, reject);
    }, function (cause) {
      reject(new ValidationError(rule, value, cause));
    });
  } else {
    resolve(value);
  }
}

function v8n() {
  return proxyContext(new Context());
}

// Custom rules
var customRules = {};

v8n.extend = function (newRules) {
  Object.assign(customRules, newRules);
};

v8n.clearCustomRules = function () {
  customRules = {};
};

function proxyContext(context) {
  return new Proxy(context, {
    get: function get(obj, prop) {
      if (prop in obj) {
        return obj[prop];
      }

      var newContext = proxyContext(context._clone());

      if (prop in availableModifiers) {
        return newContext._applyModifier(availableModifiers[prop], prop);
      }
      if (prop in customRules) {
        return newContext._applyRule(customRules[prop], prop);
      }
      if (prop in availableRules) {
        return newContext._applyRule(availableRules[prop], prop);
      }
    }
  });
}

var availableModifiers = {
  not: {
    simple: function simple(fn) {
      return function (value) {
        return !fn(value);
      };
    },
    async: function async(fn) {
      return function (value) {
        return Promise.resolve(fn(value)).then(function (result) {
          return !result;
        });
      };
    }
  },

  some: {
    simple: function simple(fn) {
      return function (value) {
        return split(value).some(fn);
      };
    },
    async: function async(fn) {
      return function (value) {
        return Promise.all(split(value).map(fn)).then(function (result) {
          return result.some(Boolean);
        });
      };
    }
  },

  every: {
    simple: function simple(fn) {
      return function (value) {
        return split(value).every(fn);
      };
    },
    async: function async(fn) {
      return function (value) {
        return Promise.all(split(value).map(fn)).then(function (result) {
          return result.every(Boolean);
        });
      };
    }
  }
};

function split(value) {
  if (typeof value === "string") {
    return value.split("");
  }
  return value;
}

var availableRules = {
  pattern: testPattern,

  // Value

  equal: function equal(expected) {
    return function (value) {
      return value == expected;
    };
  },
  exact: function exact(expected) {
    return function (value) {
      return value === expected;
    };
  },


  // Types

  string: makeTestType("string"),

  number: function number() {
    var allowInfinite = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : true;
    return function (value) {
      return typeof value === "number" && (allowInfinite || isFinite(value));
    };
  },

  boolean: makeTestType("boolean"),

  undefined: makeTestType("undefined"),

  null: makeTestType("null"),

  array: makeTestType("array"),

  object: makeTestType("object"),

  // Pattern

  lowercase: makeTestPattern(/^([a-z]+\s*)+$/),

  uppercase: makeTestPattern(/^([A-Z]+\s*)+$/),

  vowel: makeTestPattern(/^[aeiou]+$/i),

  consonant: makeTestPattern(/^(?=[^aeiou])([a-z]+)$/i),

  // Value at

  first: makeTestValueAt(0),

  last: makeTestValueAt(-1),

  // Length

  empty: makeTestLength(0, 0),

  length: makeTestLength(),

  minLength: makeTestLength(undefined, Infinity),

  maxLength: makeTestLength(-Infinity),

  // Range

  negative: makeTestRange(-Infinity, 0),

  positive: makeTestRange(-1, Infinity),

  between: makeTestRange(undefined, undefined, true, true),

  range: makeTestRange(undefined, undefined, true, true),

  lessThan: makeTestRange(-Infinity, undefined, true),

  lessThanOrEqual: makeTestRange(-Infinity, undefined, true, true),

  greaterThan: makeTestRange(undefined, Infinity, false, true),

  greaterThanOrEqual: makeTestRange(undefined, Infinity, true, true),

  // Divisible

  even: makeTestDivisible(2, true),

  odd: makeTestDivisible(2, false),

  includes: function includes(expected) {
    return testIncludes(expected);
  },


  integer: function integer() {
    return testInteger();
  },

  schema: function schema(_schema) {
    return testSchema(_schema);
  },

  // branching

  passesAnyOf: function passesAnyOf() {
    for (var _len = arguments.length, validations = Array(_len), _key = 0; _key < _len; _key++) {
      validations[_key] = arguments[_key];
    }

    return function (value) {
      return validations.some(function (validation) {
        return validation.test(value);
      });
    };
  },

  optional: function optional(validation) {
    return function (value) {
      if (value === undefined || value === null) return true;
      validation.check(value);
      return true;
    };
  }
};

function testPattern(pattern) {
  return function (value) {
    return pattern.test(value);
  };
}

function makeTestPattern(pattern) {
  return function () {
    return testPattern(pattern);
  };
}

function makeTestType(type) {
  return function () {
    return function (value) {
      return typeof value === type || value === null && type === "null" || Array.isArray(value) && type === "array";
    };
  };
}

function makeTestValueAt(index) {
  return function (expected) {
    return function (value) {
      var i = index < 0 ? value.length + index : index;
      return value[i] == expected;
    };
  };
}

function makeTestLength(overriddenMin, overriddenMax) {
  return function (min, max) {
    return function (value) {
      return value.length >= firstDefined([overriddenMin, min]) && value.length <= firstDefined([overriddenMax, max, min]);
    };
  };
}

function makeTestRange(overriddenLower, overriddenUpper, inclusiveLower, inclusiveUpper) {
  return function (lower, upper) {
    return function (value) {
      upper = firstDefined([overriddenUpper, upper, lower]);
      lower = firstDefined([overriddenLower, lower]);
      return (inclusiveLower ? value >= lower : value > lower) && (inclusiveUpper ? value <= upper : value < upper);
    };
  };
}

function makeTestDivisible(by, expected) {
  return function () {
    return function (value) {
      return value % by === 0 === expected;
    };
  };
}

function testIncludes(expected) {
  return function (value) {
    return value.indexOf(expected) !== -1;
  };
}

function testInteger() {
  var isInteger = Number.isInteger || isIntegerPolyfill;
  return function (value) {
    return isInteger(value);
  };
}

function isIntegerPolyfill(value) {
  return typeof value === "number" && isFinite(value) && Math.floor(value) === value;
}

function testSchema(schema) {
  return function (value) {
    var causes = [];
    Object.keys(schema).forEach(function (key) {
      var nestedValidation = schema[key];
      try {
        nestedValidation.check((value || {})[key]);
      } catch (ex) {
        ex.target = key;
        causes.push(ex);
      }
    });
    if (causes.length > 0) {
      throw causes;
    }
    return true;
  };
}

function firstDefined(values) {
  return values.find(function (it) {
    return it !== undefined;
  });
}

export default v8n;
